import glob
import os

import cv2


class WebCamera:
    def __init__(self, index=0):
        self.cap = cv2.VideoCapture(index)

    def get_frame(self):
        ret, frame = self.cap.read()
        return ret, frame

    def get_fps_size_len(self):
        fps = self.cap.get(cv2.CAP_PROP_FPS)
        size = (
            int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )
        length = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        return (fps, size, length)

    def isOpened(self):
        if self.cap.isOpened:
            return True
        else:
            return False

    def stop(self):
        self.cap.release()


class LoadImgData:
    def __init__(self, path=None, img_root=None):
        self.frame_index = 0
        self.frame_num = 0
        if os.path.isdir(path):
            self.img_path_list = glob.glob(os.path.join(path, "*.jpg"))
        else:
            self.img_path_list = [
                line.split()[0] for line in open(path, "r").readlines()
            ]
            if img_root:
                self.img_path_list = [
                    os.path.join(img_root, p) for p in self.img_path_list
                ]
        self.frame_num = len(self.img_path_list)

    def isOpened(self):
        if self.frame_index <= self.frame_num - 1:
            return True
        else:
            return False

    def get_frame(self):
        img_path = self.img_path_list[self.frame_index]
        frame = cv2.imread(img_path)
        # self.frame_index = min(self.frame_index + 1, self.frame_num - 1)
        self.frame_index = self.frame_index + 1
        if self.frame_index <= self.frame_num - 1:
            success = True
        else:
            success = False
        return success, frame, img_path

    def stop(self):
        exit(1)



import random

import cv2


def draw_joints(img, joint, color=None, thickness=3, radius=10):
    """
    Draw 2d hand joints on image
    :param img: image to draw
    :param joint: 2d hand joints. (21*2)
    """
    joint_id_start = [0, 0, 0, 0, 0, 1, 6, 7, 2, 9, 10, 3, 12, 13, 4, 15, 16, 5, 18, 19]
    joint_id_end = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
    ]
    red = [48, 48, 255]
    green = [48, 255, 48]
    blue = [192, 101, 21]
    yellow = [0, 204, 255]
    purple = [128, 64, 128]
    cyan = [180, 229, 255]
    joint_color_list = [
        cyan,
        green,
        blue,
        yellow,
        purple,
        red,
        green,
        green,
        green,
        blue,
        blue,
        blue,
        yellow,
        yellow,
        yellow,
        purple,
        purple,
        purple,
        red,
        red,
        red,
    ]
    bone_color_list = [
        green,
        blue,
        yellow,
        purple,
        red,
        green,
        green,
        green,
        blue,
        blue,
        blue,
        yellow,
        yellow,
        yellow,
        purple,
        purple,
        purple,
        red,
        red,
        red,
    ]
    for i in range(len(joint_id_start)):
        i1 = joint_id_start[i]
        i2 = joint_id_end[i]
        p1 = (int(joint[i1][0]), int(joint[i1][1]))
        p2 = (int(joint[i2][0]), int(joint[i2][1]))
        color_t = bone_color_list[i] if color is None else color
        cv2.line(img, p1, p2, color=color_t, thickness=thickness, lineType=cv2.LINE_AA)
    for i, j in enumerate(joint):
        p = int(j[0] * 4), int(j[1] * 4)
        color_t = joint_color_list[i] if color is None else color
        cv2.circle(
            img,
            p,
            radius=radius,
            color=[224, 224, 224],
            thickness=thickness,
            lineType=cv2.LINE_AA,
            shift=2,
        )
        cv2.circle(
            img,
            p,
            radius=radius - 4,
            color=color_t,
            thickness=thickness,
            lineType=cv2.LINE_AA,
            shift=2,
        )


def draw_one_box(img, bbox, color=None, label=None, line_thickness=None):
    """
    Draw one hand box on image
    :param img: image to draw
    :param bbox: hand bounding box. list or numpy.ndarray, xyxy
    :param color: box color
    :param label: box label
    """
    tl = (
        line_thickness or round(0.002 * (img.shape[0] + img.shape[1]) / 2) + 1
    )  # line/font thickness
    color = color or [random.randint(0, 255) for _ in range(3)]
    c1, c2 = (int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3]))
    cv2.rectangle(img, c1, c2, color, thickness=tl, lineType=cv2.LINE_AA)
    if label:
        tf = max(tl - 1, 1)  # font thickness
        t_size = cv2.getTextSize(label, 0, fontScale=tl / 3, thickness=tf)[0]
        c2 = c1[0] + t_size[0], c1[1] - t_size[1] - 3
        cv2.rectangle(img, c1, c2, color, -1, cv2.LINE_AA)  # filled
        cv2.putText(
            img,
            label,
            (c1[0], c1[1] - 2),
            0,
            tl / 3,
            [225, 255, 255],
            thickness=tf,
            lineType=cv2.LINE_AA,
        )


def draw_corner_box(img, bbox, label=None, k=30, t=4, rt=1):
    """
    Draw hand box with corner effects
    :param img: image to draw
    :param bbox: hand bounding box. list or numpy.ndarray, xyxy
    """
    colorR = (255, 0, 255)
    colorC = (0, 255, 0)
    x, y, x1, y1 = int(bbox[0]), int(bbox[1]), int(bbox[2]), int(bbox[3])

    if rt != 0:
        cv2.rectangle(img, (x, y), (x1, y1), colorR, rt)
    # Top Left  x,y
    cv2.line(img, (x, y), (x + k, y), colorC, t)
    cv2.line(img, (x, y), (x, y + k), colorC, t)
    # Top Right  x1,y
    cv2.line(img, (x1, y), (x1 - k, y), colorC, t)
    cv2.line(img, (x1, y), (x1, y + k), colorC, t)
    # Bottom Left  x,y1
    cv2.line(img, (x, y1), (x + k, y1), colorC, t)
    cv2.line(img, (x, y1), (x, y1 - k), colorC, t)
    # Bottom Right  x1,y1
    cv2.line(img, (x1, y1), (x1 - k, y1), colorC, t)
    cv2.line(img, (x1, y1), (x1, y1 - k), colorC, t)
    if label:
        cv2.putText(
            img, label, (x, y - 4), 0, 1, colorR, thickness=2, lineType=cv2.LINE_AA
        )



import time
from collections import deque, namedtuple

import numpy as np
from scipy.signal import savgol_filter


class LowPassFilter:
    def __init__(self, alpha=1.0):
        self.alpha = alpha
        self.initialized = False

    def apply(self, value):
        # Note that value can be a scalar or a numpy array
        if self.initialized:
            v = self.alpha * value + (1.0 - self.alpha) * self.stored_value
        else:
            v = value
            self.initialized = True
        self.stored_value = v
        return v

    def apply_with_alpha(self, value, alpha):
        self.alpha = alpha
        return self.apply(value)


class OneEuroFilter:
    def __init__(self, x0, dx0=0.0, min_cutoff=1.7, beta=0.3, d_cutoff=30.0, fps=None):
        """One Euro Filter for keypoints smoothing.
        Args:
            x0 (np.ndarray[K, 2]): Initialize keypoints value
            dx0 (float): 0.0
            min_cutoff (float): parameter for one euro filter
            beta (float): parameter for one euro filter
            d_cutoff (float): Input data FPS
            fps (float): Video FPS for video inference
        """

        # The parameters.
        self.data_shape = x0.shape
        self.min_cutoff = np.full(x0.shape, min_cutoff)
        self.beta = np.full(x0.shape, beta)
        self.d_cutoff = np.full(x0.shape, d_cutoff)
        # Previous values.
        self.x_prev = x0.astype(np.float32)
        self.dx_prev = np.full(x0.shape, dx0)
        self.mask_prev = np.ma.masked_where(x0 <= 0, x0)
        self.realtime = True
        if fps is None:
            # Using in realtime inference
            self.t_e = None
            self.skip_frame_factor = d_cutoff
        else:
            # fps using video inference
            self.realtime = False
            self.d_cutoff = np.full(x0.shape, float(fps))
        self.t_prev = time.time()

    def smoothing_factor(self, t_e, cutoff):
        r = 2 * np.pi * cutoff * t_e
        return r / (r + 1)

    def exponential_smoothing(self, a, x, x_prev):
        return a * x + (1 - a) * x_prev

    def update(self, x, t_e=1.0):
        """Compute the filtered signal.
        parameter (cutoff, beta) from VNect
        (http://gvv.mpi-inf.mpg.de/projects/VNect/)
        Realtime Camera fps (d_cutoff) default 30.0
        Args:
            x (np.ndarray[K, 2]): keypoints results in frame
            t_e (Optional): video skip frame count for posetrack
                evaluation
        """
        assert x.shape == self.data_shape

        t = 0
        if self.realtime:
            t = time.time()
            t_e = (t - self.t_prev) * self.skip_frame_factor
        t_e = np.full(x.shape, t_e)
        # missing keypoints mask
        mask = np.ma.masked_where(x <= 0, x)

        # The filtered derivative of the signal.
        a_d = self.smoothing_factor(t_e, self.d_cutoff)
        dx = (x - self.x_prev) / t_e
        dx_hat = self.exponential_smoothing(a_d, dx, self.dx_prev)

        # The filtered signal.
        cutoff = self.min_cutoff + self.beta * np.abs(dx_hat)
        a = self.smoothing_factor(t_e, cutoff)
        self.x_hat = self.exponential_smoothing(a, x, self.x_prev)

        # missing keypoints remove
        np.copyto(self.x_hat, -10, where=mask.mask)

        # Memorize the previous values.
        self.x_prev = self.x_hat
        self.dx_prev = dx_hat
        self.t_prev = t
        self.mask_prev = mask

    def get_results(self):
        return self.x_hat


class RelativeVelocityFilter:
    def __init__(self, window_size=5, velocity_scale=10, shape=1):
        self.window_size = window_size
        self.velocity_scale = velocity_scale
        self.last_value = np.zeros(shape)
        self.last_value_scale = np.ones(shape)
        self.last_timestamp = -1
        self.window = deque(maxlen=self.window_size)
        self.WindowElement = namedtuple("WindowElement", ["distance", "duration"])
        self.lpf = LowPassFilter()

    def get_object_scale(self, landmarks):
        lm_min = np.min(landmarks[:2], axis=1)  # min x , min y
        lm_max = np.max(landmarks[:2], axis=1)  # max x , max y
        return np.mean(lm_max - lm_min)  # average of object width and object height

    def update(self, value, timestamp=None):
        # Applies filter to the value.
        # timestamp - timestamp associated with the value (for instance,
        #             timestamp of the frame where you got value from)
        # value_scale - value scale (for instance, if your value is a distance
        #               detected on a frame, it can look same on different
        #               devices but have quite different absolute values due
        #               to different resolution, you should come up with an
        #               appropriate parameter for your particular use case)
        # value - value to filter
        value_scale = 1 / self.get_object_scale(value)
        if timestamp is None:
            timestamp = time.perf_counter()
        if self.last_timestamp == -1:
            alpha = 1.0
        else:
            distance = value * value_scale - self.last_value * self.last_value_scale
            duration = timestamp - self.last_timestamp
            cumul_distance = distance.copy()
            cumul_duration = duration
            # Define max cumulative duration assuming
            # 30 frames per second is a good frame rate, so assuming 30 values
            # per second or 1 / 30 of a second is a good duration per window element
            max_cumul_duration = (1 + len(self.window)) * 1 / 30
            for el in self.window:
                if cumul_duration + el.duration > max_cumul_duration:
                    break
                cumul_distance += el.distance
                cumul_duration += el.duration
            velocity = cumul_distance / cumul_duration
            alpha = 1 - 1 / (1 + self.velocity_scale * np.abs(velocity))
            self.window.append(self.WindowElement(distance, duration))

        self.last_value = value
        self.last_value_scale = value_scale
        self.last_timestamp = timestamp
        self.value_fi = self.lpf.apply_with_alpha(value, alpha)

    def get_results(self):
        return self.value_fi


class SavgolFilter:
    def __init__(self, queue_len=15, poly_order=3):
        self.queue_len = queue_len
        self.win_size = queue_len - 1 if queue_len % 2 == 0 else queue_len - 2
        self.poly_order = poly_order
        self.kpt_queue = deque(maxlen=queue_len)

    def update(self, x):
        self.kpt_queue.append(x)
        if len(self.kpt_queue) < self.queue_len:
            self.x_hat = x
        else:
            transKpts = np.array(self.kpt_queue).transpose(1, 2, 0)
            result = savgol_filter(transKpts, self.win_size, self.poly_order).transpose(
                2, 0, 1
            )
            self.x_hat = result[-1, :, :]

    def get_results(self):
        return self.x_hat



from distutils.core import setup

setup(
    name="handzone",
    packages=["handzone"],
    version="0.0.5",
    license="MIT",
    description="Hand Utils",
    author="cavalleria",
    author_email="cavallyb@gmail.com",
    url="https://github.com/cavalleria/handzone.git",
    install_requires=["opencv-python", "numpy", "scipy"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
